;; Generated package description from hermes.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "hermes" "0.2.0.0.20260813.54" "Emacs frontend for Hermes Agent" '((emacs "29.1") (keymap-popup "0.3.1") (websocket "1.15") (markdown-mode "2.6")) :commit "feef6c5b609d0b90cdb5216ba79c316df38a3b4a" :authors '(("Thanos Apollo" . "public@thanosapollo.org")) :maintainer '("Thanos Apollo" . "public@thanosapollo.org") :keywords '("tools" "convenience") :url "https://git.thanosapollo.org/emacs-hermes")
